var http = require('http');
var swig = require('swig');
var qs = require('querystring');
var fs = require('fs');
var sha256 = require('sha256');


/**
 * answer checking functions
 * 2 sections, 5 questions each
 */
var checkers = {
  "s1": {},
  "s2": {}
}

/**
 * "is numeric" check
 * source: http://stackoverflow.com/a/14794066
 */
function isInt(val) {
  return (!isNaN(val) && (function(x) { return (x | 0) === x; })(parseFloat(val)))
}

/**
 * simple check of closed-option radio buttons
 * the value has to be a numeric one, and be between min and max (inclusive)
 * empty string fails the test, too!
 */
function checkRadios(val, min, max) {
  if (!isInt(val)) return false;
  if (val < min) return false;
  if (val > max) return false;
  return true;
}

/**
 * checkbox check
 * one of the chechboxes has to be checked
 */
function checkCheckboxes(cbs) {
  //for c in cbs
  return true;
}

/**
 * simple text check -- we don't want nor need anything longer than 1024 characters, eh?
 */
function checkText(val) {
  if (typeof val !== "string") return false;
  if (val.length === 0) return false;
  if (val.length > 1024) return false;
  return true;
}


/**
 * just a simple check, 1 <= val <= 6
 */
checkers["s1"]["q1"] = function(post, s, q) {
  return checkRadios(post[s + '_' + q], 1, 6);
}

/**
 * 9 answers, each 1 <= val <= 5, with the last one being optional
 * if last one marked, non-empty text!
 */
checkers["s1"]["q2"] = function(post, s, q) {
  
  // these that have not been found
  var errors = Array();
  
  // 9 options, each can be equal 1 to 5
  for (var i = 1; i <= 8; i++) {
    if (!checkRadios(post[s + '_' + q + '_' + i], 1, 5))
      errors.push(s + '_' + q + '_' + i)
  }
  
  // last option requires text!
  if (checkRadios(post[s + '_' + q + '_' + 9], 1, 5)) {
    if (!checkText(post[s + '_' + q + '_' + 9 + '_text']))
      errors.push(s + '_' + q + '_' + 9 + '_text')
  }
  
  // seems we're home
  if (errors.length > 0) {
    return errors;
  } else {
    return true;
  }
}

/**
 * just a simple check, 0 <= val <= 1
 */
checkers["s1"]["q3"] = function(post, s, q) {
  return checkRadios(post[s + '_' + q], 0, 1);
}

/**
 * 6 answers, each 0 <= val <= 1, at least one required
 */
checkers["s1"]["q4"] = function(post, s, q) {
  // 6 options, each can either not exist or be equal '1'
  for (var i = 1; i <= 6; i++) {
    if (post[s + '_' + q + '_' + i] === '1')
      return true;
  }
  return false;
}

/**
 * just a simple check, 1 <= val <= 4
 */
checkers["s1"]["q5"] = function(post, s, q) {
  return checkRadios(post[s + '_' + q], 1, 4);
}

/**
 * optional value, simple check 1 <= val <= 7
 * if val == 7, check text length (but nothing more)
 */
checkers["s2"]["q1"] = function(post, s, q) {
  return ( typeof post[s + '_' + q] === 'undefined') // nothing to check/save here, at all
    || checkRadios(post[s + '_' + q], 1, 6) // one of the options is selected, and it's not the "other" option (hence, text is irrelevant)
    || ( (post[s + '_' + q] == 7) // whooops, that's the "other" option, need to check if text isn't too long!
      && ( ( post[s + '_' + q + '_7_text'] == false ) || checkText(post[s + '_' + q + '_7_text']) ) // empty? not too long otherwise?
    );
}

/**
 * optional value, simple check, 1 <= val <= 8
 */
checkers["s2"]["q2"] = function(post, s, q) {
  return (typeof post[s + '_' + q] === 'undefined') || checkRadios(post[s + '_' + q], 1, 8);
}

/**
 * optional value, simple check, 1 <= val <= 5
 */
checkers["s2"]["q3"] = function(post, s, q) {
  return (typeof post[s + '_' + q] === 'undefined') || checkRadios(post[s + '_' + q], 1, 5);
}

/**
 * optional value, simple check, 1 <= val <= 5
 */
checkers["s2"]["q4"] = function(post, s, q) {
  return (typeof post[s + '_' + q] === 'undefined') || checkRadios(post[s + '_' + q], 1, 5);
}

/**
 * optional value, simple text check
 */
checkers["s2"]["q5"] = function(post, s, q) {
  return (typeof post[s + '_' + q] === 'undefined') || (post[s + '_' + q] === '') || checkText(post[s + '_' + q]);
}


/**
 * checking the POST data
 */
function checkPostData(post) {
  // errors!
  var errors = {}

  // do we have *anything* beyond what we expect?
  var fields = [ "s1_q1", "s1_q2_1", "s1_q2_2", "s1_q2_3", "s1_q2_4", "s1_q2_5", "s1_q2_6", "s1_q2_7", "s1_q2_8", "s1_q2_9", "s1_q2_9_text", "s1_q3", "s1_q4_1", "s1_q4_2", "s1_q4_3", "s1_q4_4", "s1_q4_5", "s1_q5", "s2_q1", "s2_q1_7_text", "s2_q2", "s2_q3", "s2_q4", "s2_q5" ]
  for (f in post) {
    // rogue field? that's an error
    if (fields.indexOf(f) < 0) {
      errors[f] = false;
      delete post[f]
    }
  }
  
  // let's check the non-rogue fields, eh
  for (s in checkers) {
    for (q in checkers[s]) {

      // if not true, it's either false (a simple error in a simple question), or some additional info
      var check = checkers[s][q](post, s, q) 

      if (check !== true) {
        if (!errors[s]) errors[s] = {}
        errors[s][q] = check
        //console.log('typeof errors[' + s + '][' + q + ']: ' + typeof errors[s][q])
      }

    }
  }
  
  // return the stuff
  return errors;
}

/**
 * hashing the request data -- IP address, user-agent string
 */
function hashRequestData(request) {
  // simple enough
  return sha256(request.headers['host']
            + '|' + request.headers['cache-control']
            + '|' + request.headers['accept']
            + '|' + request.headers['user-agent']
            + '|' + request.headers['content-type']
            + '|' + request.headers['accept-encoding']
            + '|' + request.headers['accept-language']);
}


/**
 * simply output the empty survey
 * optionally with data submitted to POST
 */
function returnSurvey(response, post, errors) {
  // headers
  response.writeHead(200, {
    'Content-Type': 'text/html'
  });

  // post
  //console.log('typeof post is: ' + typeof post)
  if (typeof post === 'undefined') post = new Array()
  // errors
  //console.log('typeof errors is: ' + typeof errors)
  if (typeof errors === 'undefined') {
    errors = {}
  } /*else {
    console.log(errors)
  }*/

  // swig the survey out!
  response.end(swig.renderFile('survey.html', {
    // no errors here
    errors: errors,
    post: post
  }));
}


/**
 * handling a Not Found error
 */
function do404(response) {
  response.writeHead(404, {
    'Content-Type': 'text/html'
  });
  response.end('<h1>404 Not Found</h1>');
}


// let's roll!
var server = http.createServer(
  
    // yay, we got a request!
    function(request, response) {

      // GET? pretty simple, then, just return the survey!
      if (request.method === 'GET') {
        
        // debug
        console.log('GET! ' + request.url)

        if (request.url === '/') {
          returnSurvey(response)
        } else {
          do404(response);
        }

      // POST? get the data!
      // inspired by: http://stackoverflow.com/a/4310087
      } else if (request.method === 'POST') {
        
        // debug
        console.log('POST!');
        
        // hash of the identifying data
        var rdh = hashRequestData(request);
        var report = 'complete/' + rdh + '.json';
        
        // does the file exist?
        if (fs.existsSync(report)) {
          // wtf?
          returnSurvey(response, false, {
            'repeat': 'You seem to already have filled-in the survey!'
          });
        
        // seems to be AOK
        } else {
        
          var body = '';
          request.on('data', function (data) {
            body += data;

            // too much POST data, kill the connection!
            if (body.length > 1e6) {
              request.connection.destroy();
            }
          });
          
          // request ended?
          request.on('end', function () {

            // get the POST data
            var post = qs.parse(body);
            var errors = checkPostData(post);
           
            if (Object.keys(errors).length > 0) {
              // return the survey if errors arose
              // yay for JS not being able to easily check if an object is empty
              returnSurvey(response, post, errors);
            } else {
              // if survey has not been returned, we're home
              // add some data
              post['request_hash'] = rdh;
              post['datetime'] = Date.now();

              // write the file
              try {
                fs.writeFileSync(report, JSON.stringify(post));
                // give the user some info and thank there
                returnSurvey(response, post, errors);
              } catch (e) {
                // wtf?
                returnSurvey(response, post, {
                  save: 'saving failed!'
                });
              }
            }
          });
        }
      }
    });

server.listen(80);

console.log('listening on http://0.0.0.0:80/');
